"""{{ project_name }} - A Python package."""

__version__ = "0.1.0"
__author__ = "{{ author }}"

if __name__ == "__main__":
    import logging

    logging.basicConfig(level=logging.INFO)
    logging.info("Hello from {{ project_name }}!")
